"""Define the reconstruction process.

"""


from collections import namedtuple
from multiprocessing import Pool, cpu_count

from cml.shared.errors import (
    ReconstructionFailure)

InterimPragmatic = namedtuple("InterimPragmatic",
                              "model, predictions, learnblock")


class Metadata:
    def __init__(self, knowledge_domain, knowledge_level, identifier,
                 learnblock_indices, learnblock_features, learnblock_labels,
                 t_min, t_max, subject, aim):
        """
        subject = ANN, RF
        aim: C.2.Kme02
        """
        assert isinstance(subject, tuple)

        self.knowledge_domain = knowledge_domain
        self.knowledge_level = knowledge_level
        self.identifier = identifier
        self.learnblock_indices = learnblock_indices
        self.learnblock_features = learnblock_features
        self.learnblock_labels = learnblock_labels
        self.t_min = t_min
        self.t_max = t_max
        self.subject = subject
        self.aim = aim

    def __str__(self):
        return "\n".join([
                "{:20}:{}".format("Knowledge Domain", self.knowledge_domain),
                "{:20}:{}".format("Knowledge Level", self.knowledge_level),
                "{:20}:{}".format("Identifier", self.identifier),
                "{:20}:{}".format("Learnblock Indices", self.learnblock_indices),
                "{:20}:{}".format("Learnblock Labels", self.learnblock_labels),
                "{:20}:{}".format("Learnblock Features", self.learnblock_features),
                "{:20}:{}".format("T Min", self.t_min),
                "{:20}:{}".format("T Max", self.t_max),
                "{:20}:{}".format("Subject", self.subject),
                "{:20}:{}".format("Aim", self.aim)
            ]
        )

    def __repr__(self):
        return ", ".format([
                "{}={}".format("knowledge_domain", self.knowledge_domain),
                "{}={}".format("knowledge_level", self.knowledge_level),
                "{}={}".format("identifier", self.identifier),
                "{}={}".format("learnblock_indices", self.learnblock_indices),
                "{}={}".format("learnblock_features", self.learnblock_features),
                "{}={}".format("learnblock_labels", self.learnblock_features),
                "{}={}".format("t_min", self.t_min),
                "{}={}".format("t_max", self.t_max),
                "{}={}".format("subject", self.subject),
                "{}={}".format("aim", self.aim),
            ]
        )

    def __hash__(self):
        return hash(
            ".".join([
                self.knowledge_domain,
                str(self.knowledge_level),
                str(self.identifier)
            ])
        )

    def __eq__(self, other):
        if not isinstance(other, Metadata):
            raise TypeError

        elif isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplemented


class PragmaticMachineLearningModel:

    def __init__(self, meta, model, learnblock):
        assert isinstance(learnblock.origin, tuple)
        self.__learnblock = learnblock if learnblock.constructed_from_models() \
                                       else None
        self.meta = meta
        self.model = model
        self.origin = learnblock.origin

    def __str__(self):
        return "\n".join([
                str(self.meta),
                "{:20}: {}".format("Origin", str(self.origin))
            ]
        )

    def __repr__(self):
        return ", ".format([
                repr(self.meta),
                "{}={}".format("Origin", str(self.origin)),
                "{}=[{}]".format("model", repr(self.model))
            ]
        )

    def __hash__(self):
        return hash(self.meta)

    def __eq__(self, other):
        if isinstance(other, PragmaticMachineLearningModel):
            return hash(self.meta) == hash(other.meta)

        if isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplemented

    @property
    def domain(self):
        return self.meta.knowledge_domain

    @property
    def level(self):
        return self.meta.knowledge_level

    @property
    def identifier(self):
        return self.meta.identifier

    @property
    def uid(self):
        return "." .join([self.domain, str(self.level), str(self.identifier)])

    @identifier.setter
    def identifier(self, value):
        self.meta.identifier = value

    @property
    def min_timestamp(self):
        return self.meta.t_min

    @property
    def max_timestamp(self):
        return self.meta.t_max

    @property
    def learnblock_features(self):
        return self.meta.learnblock_features

    @property
    def learnblock_indices(self):
        return self.meta.learnblock_indices

    @property
    def learnblock_labels(self):
        return self.meta.learnblock_labels

    @property
    def subject(self):
        return self.meta.subject

    @property
    def aim(self):
        return self.meta.aim

    @property
    def accuracy(self):
        return self.model.accuracy

    @property
    def reliability(self):
        return self.model.reliability

    @reliability.setter
    def reliability(self, value):
        self.model.reliability = value

    def fusion(self, model, knowledge_level, new_identifier):
        return Metadata(knowledge_domain=self.meta.knowledge_domain,
                        knowledge_level=knowledge_level,
                        identifier=new_identifier,
                        learnblock_indices=None,
                        learnblock_features=None,
                        learnblock_labels=None,
                        t_min=min(self.meta.t_min, model.min_timestamp),
                        t_max=max(self.meta.t_max, model.max_timestamp),
                        subject=tuple(set(self.subject+model.subject)),
                        aim=tuple(set(self.meta.aim+model.aim)))

    def trained_with(self, knowledge):
        if self.__learnblock:
            return self.__learnblock

        learnblock = knowledge.get_block(self.learnblock_indices)
        preserved_features = self.learnblock_features + ["T", "Sigma", "Z"]
        for feature in learnblock.columns():
            if feature not in preserved_features:
                learnblock.drop_columns([feature])
        learnblock.set_column("Z", self.learnblock_labels)
        learnblock.subject = self.subject
        learnblock.origin = self.origin
        return learnblock


class AbstractReconstructor:
    CONCEPTUAL_KNOWLEDGE_ABBREVIATION = "C"
    PROCEDURAL_KNOWLEDGE_ABBREVIATION = "P"

    def __init__(self, mode, ml_models, settings, krippendorf_alpha):
        self.mode = mode
        self.ml_models = ml_models
        self.settings = settings
        self.model_counter = None
        self.krippendorf_alpha = krippendorf_alpha

    def reconstruct(self, *args, **kwargs):
        raise NotImplementedError

    def inject_counter(self, counter):
        self.model_counter = counter

    def to_dict(self):
        return {
            "mode": self.mode,
            "ml_models": self.ml_models,
            "settings": self.settings,
            "krippendorf_alpha": self.krippendorf_alpha
        }

    def __str__(self):
        return "\n" .join([
                "{:20}: {}".format("Mode", str(self.mode)),
                "{:20}: {}".format("ML Models", [
                    str(m) for m in self.ml_models]),
                str(self.settings)
            ]
        )

    def __repr__(self):
        return ", ".join([
                "{}={}".format("mode", repr(self.mode)),
                "{}".format(repr(self.settings)),
                "{}={}".format("ml_models", repr(self.ml_models))
            ]
        )

    def calc_intersubject_reliability(self, interims):
        predictions = [interim.predictions for interim in interims]
        if len(predictions) == 1:
            return 1.0

        reliability_alpha = self.krippendorf_alpha(predictions)
        return reliability_alpha


class SerialReconstructor(AbstractReconstructor):

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def reconstruct(self, *args):
        interims = list(self.build_interim_pragmatics(*args))
        if not any(interims):
            raise ReconstructionFailure

        reliability = self.calc_intersubject_reliability(interims)

        if reliability < self.settings.min_reliability:
            raise ReconstructionFailure

        for interim in interims:
            interim.model.reliability = reliability

        return reliability, [(interim.model, interim.learnblock)
                             for interim in interims]

    def build_interim_pragmatics(self, level, learnblock, which_models, meta):
        train, test = split(learnblock, self.settings.reliability_sample)
        train_X = train.as_numpy_array()
        train_Y = [label for label in train.get_column("Z")]
        test_X = test.as_numpy_array()

        for model in self.machine_learning_models(which_models):
            trained_model = model.train(train_X, train_Y)
            if valid_reconstructed(trained_model, self.mode, self.settings):
                predictions = trained_model.predict(test_X)
                metadata = self.get_meta(level, learnblock, trained_model, meta)
                pragmatic = PragmaticMachineLearningModel(metadata,
                                                          trained_model,
                                                          learnblock)
                yield InterimPragmatic(pragmatic, predictions, learnblock)

    def machine_learning_models(self, which_models):
        if not which_models:
            for model in self.ml_models:
                yield model
        else:
            for model in self.ml_models:
                if model.subject in which_models:
                    yield model

    def get_meta(self, level, learnblock, model, meta):
        if not meta:
            zeta = self.aim(level, learnblock)
            return Metadata(
                knowledge_domain=self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                knowledge_level=level, identifier=self.model_counter(level),
                learnblock_indices=learnblock.indices(),
                learnblock_features=learnblock.columns(),
                learnblock_labels=learnblock.get_column("Z"),
                t_min=learnblock.min_timestamp(),
                t_max=learnblock.max_timestamp(), subject=(model.subject,),
                aim=(zeta,))

        else:
            meta.learnblock_indices = learnblock.indices()
            meta.learnblock_features = learnblock.columns()
            meta.learnblock_labels = learnblock.get_column("Z")
            return meta

    def aim(self, level, learnblock):
        return ".".join([self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION, str(level),
                         learnblock.subject])


SubprocessArguments = namedtuple(
    "SubprocessArguments",
    "count, model, learnblock, level, abv, settings"
)


class ParallelReconstructor(AbstractReconstructor):

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def reconstruct(self, level, learnblock, which_models, meta):

        if not which_models:
            args = self.prepare_subprocess_args(learnblock, level)
            interims = list(self.run_background_subprocesses(args))

            if not all(interims):
                raise ReconstructionFailure

            reliability = self.calc_intersubject_reliability(interims)

            if reliability < self.settings.min_reliability:
                raise ReconstructionFailure

            for interim in interims:
                interim.model.identifier = self.model_counter(level)
                interim.model.reliability = reliability

            return reliability, [(interim.model, interim.learnblock)
                                 for interim in interims]

        else:
            return SerialReconstructor(**self.to_dict()).reconstruct(
                level,
                learnblock,
                which_models,
                meta
            )

    def prepare_subprocess_args(self, *args):
        for count, model in enumerate(self.ml_models, 1):
            yield SubprocessArguments(
                count, model, args[0], args[1],
                self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                self.settings
            )

    def run_background_subprocesses(self, args):
        from domain.reconstruction.reconstruction_parallel import reconstruct_in_parallel
        pool = Pool(cpu_count() - 2)
        interims = pool.map(reconstruct_in_parallel, args)
        pool.close()
        pool.join()
        for iterim in interims:
            if not iterim:
                continue
            yield iterim


class Reconstructor:
    def __init__(self, mode, ml_models, settings, krippendorf_alpha):
        self.state = SerialReconstructor(mode, ml_models, settings,
                                         krippendorf_alpha)

    def __str__(self):
        return str(self.state)

    def __repr__(self):
        return repr(self.state)

    def reconstruct(self, level, learnblock, which_models=None, meta=None):
        return self.state.reconstruct(level, learnblock, which_models, meta)

    def transit(self, parallel):
        if parallel:
            self.state = ParallelReconstructor(**self.state.to_dict())
        else:
            self.state = SerialReconstructor(**self.state.to_dict())

    def use_counter(self, counter):
        self.state.inject_counter(counter)

    @property
    def mode(self):
        return self.state.mode

    @mode.setter
    def mode(self, value):
        self.state.mode = value

    @property
    def min_test_accuracy(self):
        return self.state.settings.min_test_accuracy

    @min_test_accuracy.setter
    def min_test_accuracy(self, value):
        self.state.settings.min_test_accuracy = value

    @property
    def max_test_error_avg(self):
        return self.state.settings.max_test_error_avg

    @max_test_error_avg.setter
    def max_test_error_avg(self, value):
        self.state.settings.max_test_error_avg = value

    @property
    def max_test_error_max(self):
        return self.state.settings.max_test_error_max

    @max_test_error_max.setter
    def max_test_error_max(self, value):
        self.state.settings.max_test_error_max = value

    @property
    def reliability_sample(self):
        return self.state.settings.reliability_sample

    @reliability_sample.setter
    def reliability_sample(self, value):
        self.state.settings.reliability_sample = value

    @property
    def min_reliability(self):
        return self.state.settings.min_reliability

    @min_reliability.setter
    def min_reliability(self, value):
        self.state.settings.min_reliability = value

    @property
    def reduce_model_redudancy(self):
        return self.state.settings.reduce_model_redudancy

    @reduce_model_redudancy.setter
    def reduce_model_redudancy(self, value):
        self.state.settings.reduce_model_redudancy = value


def split(learnblock, reliability_sample):
    indices = learnblock.indices()
    eval_size = int(learnblock.n_rows() * reliability_sample)
    eval_idx, train_idx = indices[:eval_size], indices[eval_size+1:]
    return learnblock.new_block_from(train_idx, index=True), \
        learnblock.new_block_from(eval_idx, index=True)


def valid_reconstructed(model, knowledge_domain, settings):
    if knowledge_domain == "conceptual":
        return model.accuracy >= settings.min_test_accuracy
