"""Define adapter machine machine algorithms of sklearn.

"""


from copy import deepcopy
from collections import Counter

import sklearn
from sklearn.neighbors.kde import KernelDensity
from numpy import array, linspace, std, argsort, random


SCIKIT_CLUSTERING_TABLE = {
    sklearn.cluster.KMeans: ("n_clusters", "labels_"),
    sklearn.cluster.birch.Birch: ("n_clusters", "labels_"),
    sklearn.cluster.SpectralClustering: ("n_clusters", "labels_"),
    sklearn.cluster.FeatureAgglomeration: ("n_clusters", "labels_"),
    sklearn.cluster.AgglomerativeClustering: ("n_clusters", "labels_"),
    sklearn.cluster.DBSCAN: ("n_clusters", "labels_"),
    sklearn.cluster.OPTICS: ("n_clusters", "labels_")
}


CLUSTERING_SPECIFIC = {
    sklearn.cluster.KMeans: True,
    sklearn.cluster.DBSCAN: False,
    sklearn.cluster.SpectralClustering: True,
    sklearn.cluster.AgglomerativeClustering: True,
    sklearn.cluster.AffinityPropagation: False,
    sklearn.cluster.MeanShift: False,
    sklearn.cluster.optics_.OPTICS: False,
    sklearn.cluster.Birch: True,

}

random.seed(10)


class MachineLearningModel:

    def train(self, data, *args, **kwargs):
        raise NotImplementedError


class FilterMethod(MachineLearningModel):
    def __init__(self, model):
        self.model = model

    def __str__(self):
        return "Filtermethod: {}".format(str(self.model))

    def __repr__(self):
        return "FilterMethod({})".format(repr(self.model))

    def train(self, data, *args, **kwargs):
        self.model = self.model.fit(data)
        return self

    def reduce(self, data):
        indices = set(self.model.get_support(indices=True))
        return {i for i in range(data.n_columns(effective=True))}.difference(indices)


class EmbeddedMethod(MachineLearningModel):
    _ONE_HUNDRET_PERCENT = 100

    def __init__(self, model):
        self.model = model

    def __str__(self):
        return "EmbeddedMethod: {}".format(str(self.model))

    def __repr__(self):
        return "EmbeddedMethod({})".format(repr(self.model))

    def train(self, data, *args, **kwargs):
        labels = data.get_column_values("Z")
        self.model = self.model.fit(data, labels)
        return self

    def reduce(self):
        importance = self.model.feature_importances_
        indices = argsort(importance)[::-1]
        if (self._ONE_HUNDRET_PERCENT*std(importance))//max(importance) >= 10:
            less_relevant = indices[-1]
            return {less_relevant}
        return {}


class ConstructionClusteringMLModel(MachineLearningModel):
    def __init__(self, model, abbreviation, parallel=False):
        self.model = model
        self.abbreviation = abbreviation
        self._cluster = None
        self.parallel = parallel

    def __str__(self):
        return "\n".join([
                "{}: {}".format("Model", str(self.model)),
                "{}: {}".format("Abbreviation", str(self.abbreviation)),
           ]
        )

    def __repr__(self):
        return ",".join([
                "{}={}".format("Model", repr(self.model)),
                "{}={}".format("Abbreviation", str(self.abbreviation)),
                "{}={}".format("Parallel", str(self.parallel)),
                "{}={}".format("Cluster", str(self._cluster))
            ]
        )

    def new_model_depending_on_cluster(self, n_cluster):
        new_model = deepcopy(self)
        new_model._cluster = n_cluster
        return new_model

    def new_model(self):
        return deepcopy(self)

    def get_labels(self):
        return self.model.__getattribute__(
            SCIKIT_CLUSTERING_TABLE[type(self.model)][1]
        )

    @property
    def cluster_specific(self):
        return CLUSTERING_SPECIFIC[type(self.model)]

    @property
    def found_cluster(self):
        return len(set(self.model.labels_))

    @property
    def cluster(self):
        return self._cluster

    @cluster.setter
    def cluster(self, value):
        self.model.__setattr__(
            SCIKIT_CLUSTERING_TABLE[type(self.model)][0], value
        )
        self._cluster = value

    @property
    def cluster_sizes(self):
        labels = self.model.__getattribute__(
            SCIKIT_CLUSTERING_TABLE[type(self.model)][1]
        )
        return Counter(labels)

    def train(self, data, *args, **kwargs):
        self.model.fit(data)


class ReconstructionConceptualMLModel(MachineLearningModel):

    def __init__(self, model, abbreviation):
        self.__model = model
        self.subject = abbreviation
        self.accuracy = None
        self.reliability = None

    def train(self, data, *args, **kwargs):
        # TODO (dmt): Improve signature of this function!
        # TODO (dmt): Check if fit can handle data.
        labels = args[0]
        self.__model = self.__model.fit(data, labels)
        self.accuracy = self.__model.score(data, labels)
        return self

    def predict(self, data):
        return [i for i in self.__model.predict(data)]


class KernelDensityEstimator(MachineLearningModel):

    def __init__(self, kernel="gaussian", bandwidth=3, gridsize=256):
        self.__model = None
        self.kernel = kernel
        self.bandwidth = bandwidth
        self.gridsize = gridsize

    def train(self, data, *args, **kwargs):
        reshaped_data = array(data).reshape(-1, 1)
        if not self.__model:
            self.__model = KernelDensity(kernel=self.kernel,
                                         bandwidth=self.bandwidth)
        self.__model.fit(reshaped_data)

        return self

    def density_two(self):
        grid = linspace(0, self.gridsize)
        reshaped_grid = grid.reshape(-1, 1)
        return self.__model.score_samples(reshaped_grid)

    def density(self, data):
        reshaped_data = array(data).reshape(-1, 1)
        return self.__model.score_samples(reshaped_data)

