import json
from datetime import datetime
from os import mkdir
from os.path import join, expanduser
from enum import Enum
from functools import partial


class ConstructionInfo:
    class State(Enum):
        DISCARDED = 1
        CONSTRUCTED = 2
        PASSED = 3

    def __init__(self, state=None, subject=None, n_cluster=None):
        self.state = state
        self.subject = subject
        self.n_cluster = n_cluster


class DeconstInfo:
    def __init__(self, status=False, relationship=None, relative=None,
                 pragmatic=None, reconstructed_info=None, deleted=None,
                 inserted=None, new_learnblock=None):

        self.status = status
        self.relationship = relationship
        self.relative = relative
        self.pragmatic = pragmatic
        self.reconstructed_info = reconstructed_info if reconstructed_info else []
        self.deleted = deleted if deleted else []
        self.inserted = inserted if inserted else []
        self.new_learnblock = new_learnblock


class Logger:
    JOURNAL = {}
    CURRENT_N_BLOCK = -1
    CURRENT_LEVEL = 1
    CURRENT_LB = 0
    CURRENT_CONSTRUCTION_SUBJECT = None
    LOG_DESTINATION = None

    def __init__(self):
        self.logger = None

    def log_n_block(self):
        Logger.CURRENT_N_BLOCK += 1
        Logger.JOURNAL[Logger.CURRENT_N_BLOCK] = {}

    def log_block_generation(self, block):
        Logger.JOURNAL[Logger.CURRENT_N_BLOCK]["BLOCK"] = {
            "rows": str(block.n_rows()),
            "columns": str(block.n_columns()),
            "origin": str(block.origin),
            "LEVEL": {}
        }

    def log_level(self, level):
        Logger.CURRENT_LEVEL = level
        constructivism = Logger.JOURNAL[Logger.CURRENT_N_BLOCK]\
                                       ["BLOCK"]\
                                       ["LEVEL"]
        constructivism[Logger.CURRENT_LEVEL] = {}
        constructivism[Logger.CURRENT_LEVEL]["LEARNBLOCKS"] = []

    def log_lb_identification(self, lb):
        learnblocks = Logger.JOURNAL[Logger.CURRENT_N_BLOCK]\
                                    ["BLOCK"]\
                                    ["LEVEL"]\
                                    [Logger.CURRENT_LEVEL]\
                                    ["LEARNBLOCKS"]
        learnblocks.append({
            "rows": str(lb.n_rows()),
            "columns": str(lb.n_columns()),
            "origin": str(lb.origin),
            "relationship": str(lb.relationship),
            "CONSTRUCTION": {},
            "WINNER": {},
            "DECONSTRUCTION": {"COMPLETE": [],
                               "T SIGMA": [],
                               "Z SIGMA": []}})

        Logger.CURRENT_LB = len(learnblocks)-1

    def log_construction(self, log):
        construction = Logger.JOURNAL[Logger.CURRENT_N_BLOCK]\
                                     ["BLOCK"]\
                                     ["LEVEL"]\
                                     [Logger.CURRENT_LEVEL]\
                                     ["LEARNBLOCKS"]\
                                     [Logger.CURRENT_LB]\
                                     ["CONSTRUCTION"]
        Logger.CURRENT_CONSTRUCTION_SUBJECT = str(log.subject)+str(log.n_cluster)
        construction[Logger.CURRENT_CONSTRUCTION_SUBJECT] = {
            "status": str(log.state)
        }

    def log_feature_selection(self, lb):
        selection = Logger.JOURNAL[Logger.CURRENT_N_BLOCK]\
                                  ["BLOCK"]\
                                  ["LEVEL"]\
                                  [Logger.CURRENT_LEVEL]\
                                  ["LEARNBLOCKS"]\
                                  [Logger.CURRENT_LB]\
                                  ["CONSTRUCTION"]\
                                  [Logger.CURRENT_CONSTRUCTION_SUBJECT]

        selection["SELECTION"] = {
            "rows": str(lb.n_rows()),
            "columns": str(lb.n_columns())
        }

    def log_reconstruction(self, interims):
        reconstruction = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                       ["BLOCK"] \
                                       ["LEVEL"] \
                                       [Logger.CURRENT_LEVEL] \
                                       ["LEARNBLOCKS"] \
                                       [Logger.CURRENT_LB] \
                                       ["CONSTRUCTION"] \
                                       [Logger.CURRENT_CONSTRUCTION_SUBJECT]

        reconstruction["RECONSTRUCTION"] = []

        for interim, lb in interims:
            reconstruction["RECONSTRUCTION"].append({
                    "uid": str(interim.uid),
                    "subject": str(interim.subject),
                    "aim": str(interim.aim),
                    "accuracy": str(interim.accuracy),
                    "reliability": str(interim.reliability)
                }
            )

    def log_winner(self, winner):
        reconstruction = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                       ["BLOCK"] \
                                       ["LEVEL"] \
                                       [Logger.CURRENT_LEVEL] \
                                       ["LEARNBLOCKS"] \
                                       [Logger.CURRENT_LB]

        reconstruction["WINNER"] = {
           "subject": str(winner.subject),
           "accuracy": str(winner.accuracy),
           "reliability": str(winner.reliability),
        }

    def log_deconstruction(self, log):
        deconstruction = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                       ["BLOCK"] \
                                       ["LEVEL"] \
                                       [Logger.CURRENT_LEVEL] \
                                       ["LEARNBLOCKS"] \
                                       [Logger.CURRENT_LB]\
                                       ["DECONSTRUCTION"]

        if log.relationship == ("complete", ):
            self.log_complete_deconstruction(log)

        elif log.relationship == ("Sigma", "T"):
            self.log_t_sigma_deconstruction(log)

        elif log.relationship == ("Sigma", "Z"):
            self.log_z_sigma_deconstruction(log)
        else:
            deconstruction["failed"] = {
                "inserted": [str(i.uid) for i in log.inserted],
                "deleted": [str(i.uid) for i in log.deleted]
            }

    def log_complete_deconstruction(self, log):
        complete = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                        ["BLOCK"] \
                                        ["LEVEL"] \
                                        [Logger.CURRENT_LEVEL] \
                                        ["LEARNBLOCKS"] \
                                        [Logger.CURRENT_LB]\
                                        ["DECONSTRUCTION"]\
                                        ["COMPLETE"]
        complete.append({
            "relative": str(log.relative),
            "inserted": str(log.inserted),
            "deleted": str(log.deleted),
            "RECONSTRUCTED": []
        })

        pos = len(complete) - 1
        for recon_info in log.reconstructed_info:
            complete[pos]["RECONSTRUCTED"].append(
                {"uid": str(recon_info),
                 "subject": str(recon_info.subject),
                 "accuracy": str(recon_info.accuracy),
                 "reliability": str(recon_info.reliability)}
            )

    def log_t_sigma_deconstruction(self, log):
        t_sigma = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                ["BLOCK"] \
                                ["LEVEL"] \
                                [Logger.CURRENT_LEVEL] \
                                ["LEARNBLOCKS"] \
                                [Logger.CURRENT_LB] \
                                ["DECONSTRUCTION"] \
                                ["T SIGMA"]
        t_sigma.append({
            "relative": str(log.relative.uid),
            "LEARNBLOCK": {
                "rows": str(log.new_learnblock.n_rows()),
                "columns": str(log.new_learnblock.n_columns()),
                "min T": str(log.new_learnblock.min_timestamp()),
                "max T": str(log.new_learnblock.max_timestamp()),
                "origin": str(log.new_learnblock.origin)}}
        )

    def log_z_sigma_deconstruction(self, log):
        z_sigma = Logger.JOURNAL[Logger.CURRENT_N_BLOCK] \
                                ["BLOCK"] \
                                ["LEVEL"] \
                                [Logger.CURRENT_LEVEL] \
                                ["LEARNBLOCKS"] \
                                [Logger.CURRENT_LB] \
                                ["DECONSTRUCTION"] \
                                ["Z SIGMA"]
        z_sigma.append({
                "relative": str(log.relative.uid),
                "inserted": [str(m.uid) for m in log.inserted],
                "deleted": [str(m.uid) for m in log.deleted]
        })


class StartLogging(Logger):
    def __init__(self, func):
        self.func = func

    def __get__(self, owner, func):
        return partial(self.__call__, owner)

    def __call__(self, owner):
        self.reset_logger_members()
        self.create_log_dir(owner)
        return self.func(owner)

    def create_log_dir(self, owner):
        home_folder_path = expanduser("~")
        cml_folder = ".cml"
        current_time = datetime.now().strftime("%d%m%Y%H%M%S")
        log_dir = join(home_folder_path, cml_folder, current_time)
        mkdir(log_dir)
        owner.log_destination = log_dir
        Logger.LOG_DESTINATION = log_dir

    def reset_logger_members(self):
        Logger.JOURNAL = {}
        Logger.CURRENT_N_BLOCK = -1
        Logger.CURRENT_LEVEL = 1
        Logger.CURRENT_LB = 0
        Logger.CURRENT_CONSTRUCTION_SUBJECT = None
        Logger.LOG_DESTINATION = None


class BlockLogging(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        return partial(self.__call__, owner)

    def __call__(self, owner, block):
        pandas_block = self.func(owner, block)
        super().log_n_block()
        super().log_block_generation(pandas_block)
        super().log_level(level=1)
        return pandas_block


class FinishLogging(Logger):
    def __init__(self, func):
        self.func = func

    def __get__(self, owner, func):
        return partial(self.__call__, owner)

    def __call__(self, owner, exc_type, exc_val, exc_tb):
        self.dump_journal()
        return self.func(owner, exc_type, exc_val, exc_tb)

    def dump_journal(self):
        file = join(Logger.LOG_DESTINATION, "log.json")
        with open(file, "w") as fp:
            json.dump(super().JOURNAL, fp)


class FirstLevelLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if not self.logger:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, block):
        for level, lb in self.func(owner, block):
            yield level, lb

    def log(self, level):
        self.logger.protocol("{:#^100}".format("TIER {}".format(level)))


class HighLevelLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.__log_destination = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger

        return partial(self.__call__, owner)

    def __call__(self, owner):
        old_level = 2
        super().log_level(old_level)
        for level, block in self.func(owner):
            if level > old_level:
                super().log_level(level)
                old_level = level

            super().log_lb_identification(block)
            self.log(block)
            yield level, block

    def log_banner(self, level):
        self.logger.protocol("{:#^100}".format("TIER {}".format(level)))

    def log(self, block):
        self.logger.protocol("{:*^100}".format("Block Generation"))
        self.logger.protocol("{:<20}{}".format("Rows:", str(block.n_rows())))
        self.logger.protocol("{:<20}{}".format("Columns:", str(block.n_columns())))
        self.logger.protocol("{:<20}{}".format("Origin:", str(block.origin)))


class LearnblockLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, block):
        lb = self.func(owner, block)
        self.log(lb)
        super().log_lb_identification(lb)
        return lb

    def log(self, lb):
        self.logger.protocol("{:*^100}".format("Learnblock identification"))
        self.logger.protocol("{:<20}{}".format("Rows:", str(lb.n_rows())))
        self.logger.protocol("{:<20}{}".format("Columns:", str(lb.n_columns())))
        self.logger.protocol("{:<20}{}".format("Origin:", str(lb.origin)))
        self.logger.protocol("{:<20}{}".format("Relationship", str(lb.relationship)))


class ReducerLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, learnblock):
        reduced = self.func(owner, learnblock)
        self.log(reduced)
        self.log_feature_selection(reduced)
        return reduced

    def log(self, lb):
        self.logger.protocol("{:*^100}".format("Feature selection"))
        self.logger.protocol("{:<20}{}".format("rows:", str(lb.n_rows())))
        self.logger.protocol("{:<20}{}".format("columns:", str(lb.n_columns())))


class ConstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, learnblock):
        for info, complete_learnblock in self.func(owner, learnblock):
            super().log_construction(info)

            self.log_begin_construction()
            if info.state == ConstructionInfo.State.PASSED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "PASSED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            elif info.state == ConstructionInfo.State.DISCARDED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "DISCARDED"))
                self.logger.protocol("{: <20}{}".format("Subject:", info.subject))
                self.logger.protocol("{: <20}{}".format("Cluster:", info.n_cluster))

            elif info.state == ConstructionInfo.State.CONSTRUCTED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "CONSTRUCTED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            if complete_learnblock is None:
                continue
            else:
                yield info, complete_learnblock

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Construction"))

    def log_learnglock_info(self, learnblock, complete_learnblock):
        self.logger.protocol(
            "\t".join(["",
                       "[{:^5}{:^5}]".format(learnblock.n_rows(),
                                             learnblock.n_columns()),
                       str(complete_learnblock.min_timestamp()),
                       str(complete_learnblock.max_timestamp()),
                       "{:<20}".format(str(complete_learnblock.origin)),
                       "{:<20}".format(str(complete_learnblock.relationship)),
                       "{:<20}".format(str(complete_learnblock.subject)),
                       ]))

    def log_constructed_info(self, logging_info):
        self.logger.protocol("State: {:<20}".format("CONSTRUCTED"))
        self.logger.protocol("{:<20}".format(logging_info.subject))
        self.logger.protocol("{:<20}".format(logging_info.n_cluster))


class ReconstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, level, learnblock):
        self.log_begin_construction()
        reliability, interims = self.func(owner, level, learnblock)
        super().log_reconstruction(interims)
        self.log_reconstructed(reliability, interims)
        return reliability, interims

    def log_reconstructed(self, reliability, interims):
        for model, lb in interims:
            self.logger.protocol("{: <20}{}".format(
                "Reconstructed:", str(model.uid)))
            self.logger.protocol("{: <20}{: <15}{}".format(
                "", "Subject:", str(model.subject)))
            self.logger.protocol("{: <20}{: <15}{}".format(
                "", "Accuracy:", str(model.accuracy)))
            self.logger.protocol("{: <20}{: <15}{}".format(
                "", "Reliability:", str(model.reliability)))

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Reconstruction"))


class WinnerLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger

        return partial(self.__call__, owner)

    def __call__(self, owner, reliabilities):
        self.winner_banner()
        winner, lb = self.func(owner, reliabilities)
        self.log(winner)
        super().log_winner(winner)
        return winner, lb

    def winner_banner(self):
        self.logger.protocol("{:*^100}".format("WINNER SELECTION"))

    def log(self, winner):
        self.logger.protocol("{: <20}{}".format("Reconstructed:", str(winner.uid)))
        self.logger.protocol("{: <20}{: <15}{}".format("", "Subject:", str(winner.subject)))
        self.logger.protocol("{: <20}{: <15}{}".format("", "Accuracy:", str(winner.accuracy)))
        self.logger.protocol("{: <20}{: <15}{}".format("", "Reliability:",str(winner.reliability)))




class DeconstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, level, pragmatic, reduced, knowledge):
        self.logger.protocol("{:*^100}".format("Deconstruction"))
        for log_info in self.func(owner, level, pragmatic, reduced, knowledge):
            self.log(log_info)
            super().log_deconstruction(log_info)

    def log(self, info):
        if info.relationship:
            self.logger.protocol("{:+^100}".format(str(info.relationship)))
            self.logger.protocol("{: <20}{}".format("Pragmatic:", str(info.pragmatic.uid)))
            self.logger.protocol("{: <20}{}".format("Relative:", str(info.relative.uid)))

            for r_info in info.reconstructed_info:
                self.log_reconstruction(r_info)

        inserted = ", ".join([str(i.uid) for i in info.inserted])
        deleted = ", ".join([str(i.uid) for i in info.deleted])
        self.logger.protocol("{: <20}{}".format("Inserted:", inserted))
        self.logger.protocol("{: <20}{}".format("Deleted:", deleted))

    def log_reconstruction(self, info):
        self.logger.protocol("{: <20}{}".format("Reconstructed:", str(info.uid)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Subject:", str(info.subject)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Accuracy:", str(info.accuracy)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Reliability:", str(info.reliability)))

    def t_sigma_log(self, info):
        self.logger.protocol("{:+^100}".format(str(info.logs.relationship)))
        self.logger.protocol("{: <20}{}".format("Pragmatic:", str(info.logs.pragmatic.uid)))
        self.logger.protocol("{: <20}{}".format("Relative:", str(info.relative.uid)))
        if info.status:
            self.logger.protocol("{: <20}[{:^5}{:^5}]".format("Dimension: ", str(info.new_learnblock.n_rows()), str(info.new_learnblock.n_columns())))
            self.logger.protocol("{: <20}{}".format("Min T: ", info.new_learnblock.min_timestamp()))
            self.logger.protocol("{: <20}{}".format("Max T: ", info.new_learnblock.max_timestamp()))
            self.logger.protocol("{: <20}{}".format("Origin: ", str(info.new_learnblock.origin)))
            self.logger.protocol("{: <20}{}".format("Subject:", str(info.new_learnblock.subject)))
