from typing import List, Tuple

from cml.usecases.usecase import Usecase
from cml.domain import (
    Preprocessor,
    LearnblockIdentifier,
    DataSource,
    FeatureSelector,
    Reconstructor,
    Constructor,
    Deconstructor,
    ClusterFinder)
from cml.domain.knowledge import (
    RelativeFinder)


class PreprocessUsecase(Usecase):
    def process(self, request):
        # Preprocess the data
        data = request.source_adapter.read_csv(request.general_settings.input_file)
        preprocesser = Preprocessor(request.preprocessing_settings)
        cleaned_data = preprocesser.clean(data)

        # Initialize learnblock identifier
        density_estimator = request.density_estimator_interface()
        # relative_extrema_finder = request.relative_extrema_interface
        settings = request.block_processing_settings
        learnblock_identifier = LearnblockIdentifier(settings,
                                                     density_estimator)
        return DataSource(cleaned_data, learnblock_identifier, settings)


class FeatureSelectionUsecase(Usecase):
    def process(self, request):
        return FeatureSelector(request.filter_method,
                               request.embedded_method,
                               request.settings)


class ReconstructionUsecase(Usecase):
    def process(self, request):
        models = create_models(request.interface, request.algorithms)
        return Reconstructor(request.mode, models, request.settings,
                             request.krippendorf)


class ConstructionUsecase(Usecase):
    def process(self, request):
        models = create_models(request.interface, request.algorithms)
        constructor = Constructor(request.mode, models, request.settings)
        constructor.construction_type = request.mode
        return constructor


class DeconstructionUsecase(Usecase):
    def process(self, request):
        relative_finder = RelativeFinder(request.general_settings.highest_level)
        cluster_finder = ClusterFinder(request.density_estimator,
                                       request.block_processing_settings)
        return Deconstructor(request.mode,
                             request.general_settings,
                             request.deconstruction_settings,
                             request.reconstructor,
                             request.build_new_learnblock,
                             relative_finder,
                             cluster_finder)


def create_models(interface, algorithms):
    models = []
    for abbreviation, algorithm in algorithms:
        model = interface(algorithm, abbreviation)
        models.append(model)
    return models

